/* -*-c-*-
 *
 * $Id: mptext.c,v 1.4 1999/12/22 15:56:56 mdw Exp $
 *
 * Textual representation of multiprecision numbers
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: mptext.c,v $
 * Revision 1.4  1999/12/22 15:56:56  mdw
 * Use clever recursive algorithm for writing numbers out.
 *
 * Revision 1.3  1999/12/10 23:23:26  mdw
 * Allocate slightly less memory.
 *
 * Revision 1.2  1999/11/20 22:24:15  mdw
 * Use function versions of MPX_UMULN and MPX_UADDN.
 *
 * Revision 1.1  1999/11/17 18:02:16  mdw
 * New multiprecision integer arithmetic suite.
 *
 */

/*----- Header files ------------------------------------------------------*/

#include <ctype.h>
#include <stdio.h>

#include "mp.h"
#include "mptext.h"
#include "paranoia.h"

/*----- Main code ---------------------------------------------------------*/

/* --- @mp_read@ --- *
 *
 * Arguments:	@mp *m@ = destination multiprecision number
 *		@int radix@ = base to assume for data (or zero to guess)
 *		@const mptext_ops *ops@ = pointer to operations block
 *		@void *p@ = data for the operations block
 *
 * Returns:	The integer read, or zero if it didn't work.
 *
 * Use:		Reads an integer from some source.  If the @radix@ is
 *		specified, the number is assumed to be given in that radix,
 *		with the letters `a' (either upper- or lower-case) upwards
 *		standing for digits greater than 9.  Otherwise, base 10 is
 *		assumed unless the number starts with `0' (octal), `0x' (hex)
 *		or `nnn_' (base `nnn').  An arbitrary amount of whitespace
 *		before the number is ignored.
 */

mp *mp_read(mp *m, int radix, const mptext_ops *ops, void *p)
{
  int r;
  int ch;
  unsigned f = 0;

  enum {
    f_neg = 1u,
    f_ok = 2u
  };

  /* --- Initialize the destination number --- */

  MP_MODIFY(m, 4);
  m->vl = m->v;
  m->f &= ~MP_UNDEF;

  /* --- Read an initial character --- */

  ch = ops->get(p);
  while (isspace(ch))
    ch = ops->get(p);

  /* --- Handle an initial sign --- */

  if (ch == '-') {
    f |= f_neg;
    ch = ops->get(p);
    while (isspace(ch))
      ch = ops->get(p);
  }

  /* --- If the radix is zero, look for leading zeros --- */

  if (radix)
    r = -1;
  else if (ch != '0') {
    radix = 10;
    r = 0;
  } else {
    ch = ops->get(p);
    if (ch == 'x') {
      ch = ops->get(p);
      radix = 16;
    } else {
      radix = 8;
      f |= f_ok;
    }
    r = -1;
  }

  /* --- Time to start --- */

  for (;; ch = ops->get(p)) {
    int x;

    /* --- An underscore indicates a numbered base --- */

    if (ch == '_' && r > 0 && r <= 36) {
      radix = r;
      m->vl = m->v;
      r = -1;
      f &= ~f_ok;
      continue;
    }

    /* --- Check that the character is a digit and in range --- */

    if (!isalnum(ch))
      break;
    if (ch >= '0' && ch <= '9')
      x = ch - '0';
    else {
      ch = tolower(ch);
      if (ch >= 'a' && ch <= 'z')	/* ASCII dependent! */
	x = ch - 'a' + 10;
      else
	break;
    }

    /* --- Sort out what to do with the character --- */

    if (x >= 10 && r >= 0)
      r = -1;
    if (x >= radix)
      break;

    if (r >= 0)
      r = r * 10 + x;

    /* --- Stick the character on the end of my integer --- */

    mp_ensure(m, MP_LEN(m) + 1);
    mpx_umuln(m->v, m->vl, m->v, m->vl - 1, radix);
    mpx_uaddn(m->v, m->vl, x);
    mp_shrink(m);
    f |= f_ok;
  }

  ops->unget(ch, p);

  /* --- Bail out if the number was bad --- */

  if (!(f & f_ok)) {
    MP_DROP(m);
    return (0);
  }

  /* --- Set the sign and return --- */

  m->f = 0;
  if (f & f_neg)
    m->f |= MP_NEG;
  return (m);
}

/* --- @mp_write@ --- *
 *
 * Arguments:	@mp *m@ = pointer to a multi-precision integer
 *		@int radix@ = radix to use when writing the number out
 *		@const mptext_ops *ops@ = pointer to an operations block
 *		@void *p@ = data for the operations block
 *
 * Returns:	Zero if it worked, nonzero otherwise.
 *
 * Use:		Writes a large integer in textual form.
 */

/* --- Simple case --- *
 *
 * Use a fixed-sized buffer and the simple single-precision division
 * algorithm to pick off low-order digits.  Put each digit in a buffer,
 * working backwards from the end.  If the buffer becomes full, recurse to
 * get another one.  Ensure that there are at least @z@ digits by writing
 * leading zeroes if there aren't enough real digits.
 */

static int simple(mp *m, int radix, unsigned z,
		  const mptext_ops *ops, void *p)
{
  int rc = 0;
  char buf[64];
  unsigned i = sizeof(buf);

  do {
    int ch;
    mpw x;

    x = mpx_udivn(m->v, m->vl, m->v, m->vl, radix);
    MP_SHRINK(m);
    if (x < 10)
      ch = '0' + x;
    else
      ch = 'a' + x - 10;
    buf[--i] = ch;
    if (z)
      z--;
  } while (i && MP_LEN(m));

  if (MP_LEN(m))
    rc = simple(m, radix, z, ops, p);
  else {
    static const char zero[32] = "00000000000000000000000000000000";
    while (!rc && z >= sizeof(zero)) {
      rc = ops->put(zero, sizeof(zero), p);
      z -= sizeof(zero);
    }
    if (!rc && z)
      rc = ops->put(zero, z, p);
  }
  if (!rc)
    ops->put(buf + i, sizeof(buf) - i, p);
  if (m->f & MP_BURN)
    BURN(buf);
  return (rc);
}

/* --- Complicated case --- *
 *
 * If the number is small, fall back to the simple case above.  Otherwise
 * divide and take remainder by current large power of the radix, and emit
 * each separately.  Don't emit a zero quotient.  Be very careful about
 * leading zeroes on the remainder part, because they're deeply significant.
 */

static int complicated(mp *m, int radix, mp **pr, unsigned i, unsigned z,
		       const mptext_ops *ops, void *p)
{
  int rc = 0;
  mp *q = MP_NEW;
  unsigned d = 1 << i;

  if (MP_LEN(m) < 8)
    return (simple(m, radix, z, ops, p));

  mp_div(&q, &m, m, pr[i]);
  if (!MP_LEN(q))
    d = z;
  else {
    if (z > d)
      z -= d;
    else
      z = 0;
    rc = complicated(q, radix, pr, i - 1, z, ops, p);
  }
  if (!rc)
    rc = complicated(m, radix, pr, i - 1, d, ops, p);
  mp_drop(q);
  return (rc);
}

/* --- Main driver code --- */

int mp_write(mp *m, int radix, const mptext_ops *ops, void *p)
{
  int rc;

  /* --- Set various things up --- */

  m = MP_COPY(m);
  MP_SPLIT(m);

  /* --- If the number is negative, sort that out --- */

  if (m->f & MP_NEG) {
    if (ops->put("-", 1, p))
      return (EOF);
  }

  /* --- If the number is small, do it the easy way --- */

  if (MP_LEN(m) < 8)
    rc = simple(m, radix, 0, ops, p);

  /* --- Use a clever algorithm --- *
   *
   * Square the radix repeatedly, remembering old results, until I get
   * something more than half the size of the number @m@.  Use this to divide
   * the number: the quotient and remainder will be approximately the same
   * size, and I'll have split them on a digit boundary, so I can just emit
   * the quotient and remainder recursively, in order.
   *
   * The array size copes with the largest number possibly representable on
   * the host machine.  Such a large number shouldn't ever arise in real use.
   */

  else {
    mp *pr[CHAR_BIT * sizeof(size_t)];
    size_t target = MP_LEN(m) / 2;
    unsigned i = 0;
    mp *z = mp_create(1);

    /* --- Set up the exponent table --- */

    z->v[0] = radix;
    z->f = 0;
    for (;;) {
      assert(((void)"Number is too unimaginably huge",
	      i < sizeof(pr) / sizeof(pr[0])));
      pr[i++] = z;
      if (MP_LEN(z) > target)
	break;
      z = mp_sqr(MP_NEW, z);
    }

    /* --- Write out the answer --- */

    rc = complicated(m, radix, pr, i - 1, 0, ops, p);

    /* --- Tidy away the array --- */

    while (i > 0)
      mp_drop(pr[--i]);
  }

  /* --- Tidying up code --- */

  MP_DROP(m);
  return (rc);
}

/*----- Test rig ----------------------------------------------------------*/

#ifdef TEST_RIG

#include <mLib/testrig.h>

static int verify(dstr *v)
{
  int ok = 1;
  int ib = *(int *)v[0].buf, ob = *(int *)v[2].buf;
  dstr d = DSTR_INIT;
  mp *m = mp_readdstr(MP_NEW, &v[1], 0, ib);
  if (m) {
    if (!ob) {
      fprintf(stderr, "*** unexpected successful parse\n"
	              "*** input [%i] = %s\n",
	      ib, v[1].buf);
      mp_writedstr(m, &d, 10);
      fprintf(stderr, "*** (value = %s)\n", d.buf);
      ok = 0;
    } else {
      mp_writedstr(m, &d, ob);
      if (d.len != v[3].len || memcmp(d.buf, v[3].buf, d.len) != 0) {
	fprintf(stderr, "*** failed read or write\n"
		        "*** input [%i]    = %s\n"
		        "*** output [%i]   = %s\n"
		        "*** expected [%i] = %s\n",
		ib, v[1].buf, ob, d.buf, ob, v[3].buf);
	ok = 0;
      }
    }
    mp_drop(m);
  } else {
    if (ob) {
      fprintf(stderr, "*** unexpected parse failure\n"
	              "*** input [%i] = %s\n"
	              "*** expected [%i] = %s\n",
	      ib, v[1].buf, ob, v[3].buf);
      ok = 0;
    }
  }

  dstr_destroy(&d);
  assert(mparena_count(MPARENA_GLOBAL) == 0);
  return (ok);
}

static test_chunk tests[] = {
  { "mptext", verify,
    { &type_int, &type_string, &type_int, &type_string, 0 } },
  { 0, 0, { 0 } }
};

int main(int argc, char *argv[])
{
  sub_init();
  test_run(argc, argv, tests, SRCDIR "/tests/mptext");
  return (0);
}

#endif

/*----- That's all, folks -------------------------------------------------*/
